import time
from typing import Literal

import graphblas as gb
import numpy as np
import pandas as pd
from scipy.sparse import csr_array
from sympy import nextprime


def get_prime_map_from_rel(
    list_of_rels: list,
    starting_value: int = 2,
    spacing_strategy: str = "step_10",
    add_inverse_edges: bool = False,
) -> tuple[dict, dict]:
    """
    Helper function that given a list of relations returns the mappings to and from the
    prime numbers used.

    Different strategies to map the numbers are available.
    "step_X", increases the step between two prime numbers by adding X to the current prime
    "factor_X", increases the step between two prime numbers by multiplying the current prime with X
    "natural", increases by 1 starting from 2
    "constant", return the same as step_10 but the values are changed to constants in create_pam_matrices

    Args:
        list_of_rels (list): iterable, contains a list of the relations that need to be mapped.
        starting_value (int, optional): Starting value of the primes. Defaults to 2.
        spacing_strategy (str, optional):  Spacing strategy for the primes. Defaults to "step_10".

    Returns:
        rel2prime: dict, relation to prime dictionary e.g. {"rel1":2}.
        prime2rel: dict, prime to relation dictionary e.g. {2:"rel1"}.
    """

    list_of_rels = [str(relid) for relid in list_of_rels]

    # Initialize dicts
    rel2prime = {}
    prime2rel = {}
    # Starting value for finding the next prime
    current_int = starting_value
    # Map each relation id to the next available prime according to the strategy used
    if spacing_strategy == "natural":
        c = 1
        for relid in list_of_rels:
            rel2prime[relid] = c
            prime2rel[c] = relid
            c += 1
    else:
        if spacing_strategy == "constant":
            spacing_strategy = "step_10"
        for relid in list_of_rels:
            cur_prime = int(nextprime(current_int))  # type: ignore
            rel2prime[relid] = cur_prime
            prime2rel[cur_prime] = relid
            if "step" in spacing_strategy:
                step = float(spacing_strategy.split("_")[1])
                current_int = cur_prime + step
            elif "factor" in spacing_strategy:
                factor = float(spacing_strategy.split("_")[1])
                current_int = cur_prime * factor
            else:
                raise NotImplementedError(
                    f"Spacing strategy : {spacing_strategy}  not understood!"
                )
    return rel2prime, prime2rel


def create_pam_matrices(
    df_train: pd.DataFrame,
    max_order: int = 5,
    use_log: bool = True,
    method: Literal["plus_times", "plus_plus", "constant"] = "plus_times",
    spacing_strategy: str = "step_10",
    eliminate_diagonal: bool = False,
    break_with_sparsity_threshold: float = -1,
    check_error_lossless: bool = True,
) -> tuple[csr_array, list[csr_array], dict, dict, bool]:

    """Helper function that creates the pam matrices.

    Args:
        df_train (pd.DataFrame): The triples in the form of a pd.DataFrame with columns
        (head, rel, tail).

        max_order (int, optional): The maximum order for the PAMs (i.e. the k-hops).
        Defaults to 5.

        use_log (bool, optional): Whether to use log of primes for numerical stability.
        Defaults to True.

        method (Literal["plus_times", "plus_plus", "constant"], optional):  Method of multiplication.
        - "plus_times": Generic matrix multiplication GrapBLAS multiplication.
        - "plus_plus": Matrix multiplication using a plus_plus semiring using GraphBLAS. Use with use_log=True recommended.
        - "constant": Replace all values in the 1-hop matrix with ones. For benchmarking purposes.
        Defaults to "plus_times".

        spacing_strategy (str, optional): The spacing strategy as mentioned in get_prime_map_from_rel.
        Defaults to "step_10".

        break_with_sparsity_threshold (int, optional): The percentage of sparsity that is not accepted.
        If one of the k-hop PAMs has lower sparsity we break the calculations and do not include it
        in the returned matrices list.
        Defaults to "step_10".

        eliminate_diagonal (bool, optional): Whether to zero-out the diagonal in each k-hop.
        (This essentially removes cyclic paths from being propagated).
        Defaults to False.

        check_error_lossless (bool, optional): Whether to check for int overflow for the loss-less matrix.
        Defaults to True.

    Returns:
        tuple[csr_matrix, list[csr_matrix], dict, dict, bool]: The first argument is the lossless 1-hop PAM with products.
        The second is a list of the lossy PAMs powers, the third argument is the node2id dictionary and
        the fourth argument is the relation to id dictionary. The fifth argument is weather the matrix creation process
        was broken due to sparsity.
    """

    # Number of unique rels and nodes
    print(df_train)
    unique_rels = sorted(list(df_train["rel"].unique()))

    unique_nodes = sorted(
        set(df_train["head"].values.tolist() + df_train["tail"].values.tolist())  # type: ignore
    )
    print(
        f"# of unique rels: {len(unique_rels)} \t | # of unique nodes: {len(unique_nodes)}"
    )

    node2id = {}
    id2node = {}
    for i, node in enumerate(unique_nodes):
        node2id[node] = i
        id2node[i] = node

    time_s = time.time()

    # Map the relations to primes
    rel2id, id2rel = get_prime_map_from_rel(
        unique_rels,
        starting_value=2,
        spacing_strategy=spacing_strategy,
    )

    # Map all node and rel values to the corresponding numerical ones.
    df_train["rel_mapped"] = df_train["rel"].map(rel2id)
    df_train["head_mapped"] = df_train["head"].map(node2id)
    df_train["tail_mapped"] = df_train["tail"].map(node2id)

    # Create the lossless representation (with product). This may lead to overflows due to the product.
    aggregated_df_lossless = (
        df_train.groupby(["head_mapped", "tail_mapped"])["rel_mapped"]
        .aggregate(np.prod)
        .reset_index()
    )

    # If we don't case about the loss-less scenario we can omit this error.
    if check_error_lossless:
        if any(aggregated_df_lossless["rel_mapped"].values > np.iinfo(np.int64).max):
            raise OverflowError(
                f"You have overflowing due to large prime number products in 1-hop PAM. Please lower the spacing strategy (lowest is 'step_1'), current is {spacing_strategy}"
            )

    # Construct the "lossless" 1-hop array (with products of primes probably)
    pam_1hop_lossless = csr_array(
        (
            aggregated_df_lossless["rel_mapped"],
            (
                aggregated_df_lossless["head_mapped"],
                aggregated_df_lossless["tail_mapped"],
            ),
        ),
        shape=(len(unique_nodes), len(unique_nodes)),
        dtype=np.int64,
    )

    # If use log, will need to re-map the values
    if use_log:
        print(f"Will map to logs!")
        id2rel = {}
        for k, v in rel2id.items():
            rel2id[k] = np.log(v)
            id2rel[np.log(v)] = k

    df_train["rel_mapped"] = df_train["rel"].map(rel2id)

    # Create the lossy 1-hop with log-sum-sum
    aggregated_df = (
        df_train.groupby(["head_mapped", "tail_mapped"])["rel_mapped"]
        .aggregate(np.sum)
        .reset_index()
    )
    pam_1hop_lossy = csr_array(
        (
            aggregated_df["rel_mapped"],
            (aggregated_df["head_mapped"], aggregated_df["tail_mapped"]),
        ),
        shape=(len(unique_nodes), len(unique_nodes)),
        dtype=np.float64,
    )

    # Map everything to 1. This is only for testing scenarios.
    if spacing_strategy == "constant":
        pam_1hop_lossy.data = np.ones_like(pam_1hop_lossy.data)

    # Create the GraphBLAS equivalent
    A_gb = gb.io.from_scipy_sparse(pam_1hop_lossy)

    # Generate the PAM^k matrices
    pam_powers = [pam_1hop_lossy]
    pam_power_gb = [A_gb]
    broke_cause_of_sparsity = False
    for ii in range(1, max_order):
        print(f"Hop {ii + 1}")
        cur_previous_power = pam_power_gb[-1].dup()
        if eliminate_diagonal:
            cur_previous_power.setdiag(0)
        updated_power_gb = cur_previous_power.mxm(A_gb, method).new()

        updated_power = gb.io.to_scipy_sparse(updated_power_gb)

        sparsity = get_sparsity(updated_power)
        print(f"Sparsity {ii + 1}-hop: {sparsity:.2f} %")
        if sparsity < 100 * break_with_sparsity_threshold and ii > 1:
            print(
                f"Stopped at {ii + 1} hops due to non-sparse matrix.. Current sparsity {sparsity:.2f} % < {break_with_sparsity_threshold}"
            )
            broke_cause_of_sparsity = True
            break
        pam_powers.append(updated_power)
        pam_power_gb.append(updated_power_gb)

    return pam_1hop_lossless, pam_powers, node2id, rel2id, broke_cause_of_sparsity


def get_sparsity(A: csr_array) -> float:
    """Calculate sparsity % of scipy sparse matrix.
    Args:
        A (scipy.sparse): Scipy sparse matrix
    Returns:
        (float)): Sparsity as a float
    """

    return 100 * (1 - A.nnz / (A.shape[0] ** 2))


if __name__ == "__main__":
    from data_loading import load_csv

    path = "../data/dummy_data/train.txt"

    df_train_orig, df_train = load_csv(path, add_inverse_edges="YES")
    # print(df_train_orig)
    (
        pam_lossles,
        pam_powers,
        node2id,
        rel2id,
        broke_cause_of_sparsity,
    ) = create_pam_matrices(
        df_train,
        max_order=2,
        method="plus_times",
        use_log=False,
        spacing_strategy="step_1",
    )
    print(rel2id)
    node_names = list(node2id.keys())
    pam_2 = pd.DataFrame(pam_powers[1].todense(), columns=node_names)
    pam_2.index = node_names  # type:ignore
    print("PAM@2")
    print(pam_2)



import numpy as np
import pandas as pd
import tqdm
from scipy.sparse import csr_array

from utils import get_primefactors


def create_ruleset(
    pam_1hop_lossless: csr_array,
    pam_powers: list[csr_array],
    use_log: bool,
    max_num_hops: int = -1,
    laplace_smoothing_alpha: float = 1,
) -> pd.DataFrame:
    """
    Simple function to generate rules/metapaths based on look-up procedure
    of non-zero overlapping elements in the k-hop matrices and the original
    graph. Currently only "same-cell" rules are generated. These means
    that only cyclic path rules of the form
    r1(A,B) ^ r2(B,C) ^ r3(C,D) => r_head(A,D) can be created.
    For more details see:
    https://arxiv.org/abs/2305.10467

    Args:
        pam_powers (csr_array): Lossless 1-hop
        pam_powers (list[csr_array]): List of the k-hop pams.
        use_log (bool): Whether the primes are logarithmed first.
        max_num_hops (int): Maximun hops to take into account when creating rules.
        Range in [1, len(pam_powers) + 1]
        laplace_smoothing_alpha (int): Laplace smoothing for confidence of rules

    Returns:
        pd.DataFrame: DataFrame with rules and details
    """

    # This will hold the whole rule-set
    all_rules = []
    # Iterate over all k-hops (including the 1-hop power)
    if max_num_hops == -1:
        max_num_hops = len(pam_powers) + 1
    k_hop_pams_to_mine = [pam_1hop_lossless] + pam_powers[1:]
    k_hop_pams_to_mine = k_hop_pams_to_mine[:max_num_hops]
    for k_hop_index, k_hop_pam in tqdm.tqdm(
        enumerate(k_hop_pams_to_mine), total=len(k_hop_pams_to_mine)
    ):
        # The rules that are extracted at this hop level
        cur_rules = []
        # These are the aligned non-zero elements between the 1-hop and the k-hop matrix
        # These indicate a probable rule k-hop chain -> head rel
        overlapping_nnz_rows, overlapping_nnz_cols = (
            k_hop_pam.multiply(pam_1hop_lossless)
        ).nonzero()
        # If there is actually any overlapping element
        if overlapping_nnz_rows.shape[0] > 0:
            # Keep the direct (1-hop) relations in these overlapping data
            overlapping_1hop_vals = pam_1hop_lossless[
                overlapping_nnz_rows, overlapping_nnz_cols
            ]
            # Iterarate over these direct relations
            for overlapping_element_index, overlapping_1hop_val in enumerate(
                overlapping_1hop_vals
            ):

                # Keep track of the corresponding k-hop semantic chain value
                k_hop_cell_value = k_hop_pam[
                    overlapping_nnz_rows[overlapping_element_index],
                    overlapping_nnz_cols[overlapping_element_index],
                ]
                # Factorize them in case there are more than one direct relations
                head_primes = get_primefactors(overlapping_1hop_val)
                # Iterate over these head relations that are implied by the k-hop chain
                # And create a rule for each one of them
                for head_prime in head_primes:
                    # If we are at hop 1, decompose the product of relations if available
                    if k_hop_index == 0:
                        k_hop_values = get_primefactors(k_hop_cell_value)
                        if use_log:
                            k_hop_values = [np.log(khv) for khv in k_hop_values]
                    # else use the whole k-hop chain as is
                    else:
                        k_hop_values = [k_hop_cell_value]
                    for k_hop_value in k_hop_values:
                        # This is so we do not have 7->7 type of rules from the 1-hop case
                        if head_prime != k_hop_value:

                            cur_rules.append(
                                {
                                    "head_rel": head_prime,
                                    "body_chain": k_hop_value,
                                    "num_hops": k_hop_index,
                                    "type": "same_cell",
                                }
                            )
            # If any rules were mined at this level
            if cur_rules:
                # The following process is to capture some support and coverage statistics
                # Because we may have many occurrences of the the rule
                df_rules = pd.DataFrame(cur_rules)

                grouped_by_head_and_body = (
                    df_rules.groupby(["head_rel", "type"])["body_chain"]
                    .value_counts()
                    .to_dict()
                )
                # This expresses the times each head relation is implied (at this k-hop)
                # For the rule 15=>5 at 2-hop, this counts the times that 5 is implied
                head_support = df_rules["head_rel"].value_counts().to_dict()
                # This expresses the times each body value is implied (e.g. 15)
                # For the rule 15=>5 at 2-hop, this counts the times that 15 is found in the 2-hop matrix
                if k_hop_index == 0:
                    bodies_to_use = [
                        item
                        for product in k_hop_pam.data
                        for item in get_primefactors(product)
                    ]
                    if use_log:
                        bodies_to_use = np.log(bodies_to_use)
                else:
                    bodies_to_use = k_hop_pam.data
                body_support = pd.Series(bodies_to_use).value_counts().to_dict()
                for (head_value, type_, body_value) in grouped_by_head_and_body:  # type: ignore
                    all_rules.append(
                        {
                            "head_rel": head_value,
                            "body": body_value,
                            "head_body_count": grouped_by_head_and_body[
                                (head_value, type_, body_value)
                            ],
                            "body_count": body_support[body_value],
                            "head_count": head_support[head_value],
                            "hop": k_hop_index + 1,
                            "type": type_,
                        }
                    )
    # Finally calculate some statistics, namely confidence, head_coverage
    # and the F1 score of these two
    all_rules_df = pd.DataFrame(all_rules)
    all_rules_df["conf"] = all_rules_df["head_body_count"] / (
        all_rules_df["body_count"] + laplace_smoothing_alpha
    )
    all_rules_df["head_coverage"] = all_rules_df["head_body_count"] / (
        all_rules_df["head_count"]
    )
    all_rules_df["score"] = 2 * (
        all_rules_df["conf"]
        * all_rules_df["head_coverage"]
        / (all_rules_df["head_coverage"] + all_rules_df["conf"])
    )
    all_rules_df = all_rules_df.sort_values(["score"], ascending=False)

    if any(all_rules_df["body"] < 0):
        print(
            f"Negative body values as rule patterns occured. This is usually due to numeric overflow. \
            Please consider using 'use_log=True' for this project..."
        )

    if use_log:
        all_rules_df["head_rel"] = all_rules_df["head_rel"].apply(np.log)

    return all_rules_df


if __name__ == "__main__":
    from data_loading import load_data
    from pam_creation import create_pam_matrices

    path = "../data/dummy_data"

    df_train_orig, df_train, df_eval, df_test, already_seen_triples = load_data(
        path, project_name="test", add_inverse_edges="NO", sep=",", skiprows=0
    )
    (
        pam_1hop_lossless,
        pam_powers,
        node2id,
        rel2id,
        broke_cause_of_sparsity,
    ) = create_pam_matrices(
        df_train,
        max_order=5,
        method="plus_times",
        use_log=False,
        spacing_strategy="step_1",
    )
    df_rules = create_ruleset(
        pam_1hop_lossless, pam_powers, use_log=False, max_num_hops=-1
    )
    print(df_rules)



from itertools import product
from typing import List, Union

import numba as nb
import numpy as np
import pandas as pd
import tqdm
from numba import types
from numba.typed import Dict
from scipy.sparse import csr_array
from sympy.ntheory import factorint, nextprime

from pam_creation import create_pam_matrices


def extend_paths(
    hop_k_paths: list[Union[tuple, int]], hop_1_vals: list[int]
) -> list[tuple]:
    extended_paths = []
    for comb in product(hop_k_paths, hop_1_vals):
        if isinstance(comb[0], tuple):
            res = tuple(list(comb[0]) + [comb[1]])
        else:
            res = tuple(comb)
        extended_paths.append(res)
    return extended_paths


def create_lattice_mappings(
    rel2id: dict[str, int], num_hops: int, use_log: bool, print_: bool = False
) -> dict[int, dict[str, dict]]:
    mappings = {
        0: {
            "path2prime": rel2id,
            "prime2path": dict(zip(rel2id.values(), rel2id.keys())),
        }
    }
    mappings[1] = {
        "path2prime": dict(
            zip([tuple([val]) for val in rel2id.values()], rel2id.values())
        ),
        "prime2path": dict(
            zip(rel2id.values(), [tuple([val]) for val in rel2id.values()])
        ),
    }
    hop_1_values = list(rel2id.values())
    if print_:
        iterator = tqdm.tqdm(range(2, num_hops + 1), total=num_hops - 1)
    else:
        iterator = range(2, num_hops + 1)
    for k in iterator:
        next_mapping = {}
        cur_prime = 3
        current_path_chains = (
            list(mappings[k - 1]["path2prime"].keys()) if k > 2 else hop_1_values
        )
        extended_paths = extend_paths(current_path_chains, hop_1_values)
        for ext_path in extended_paths:
            if use_log:
                value_to_use = np.log(cur_prime)
            else:
                value_to_use = cur_prime
            next_mapping[ext_path] = value_to_use
            cur_prime = nextprime(cur_prime)
        mappings[k] = {
            "path2prime": next_mapping,
            "prime2path": dict(zip(next_mapping.values(), next_mapping.keys())),
        }
    return mappings


@nb.jit(nopython=True)
def inner_direct(
    value_k_hop,
    value_1_hop,
    integer2path_k_hop,
    integer2paths_nb_1hop,
    path2prime_k_plus_1_hop,
):

    path_chains_at_k = integer2path_k_hop[value_k_hop]
    factor_last_hop = [prime for prime in integer2paths_nb_1hop[value_1_hop].flatten()]
    extended_paths_k_plus_1 = np.zeros(
        (len(path_chains_at_k) * len(factor_last_hop), len(path_chains_at_k[0]) + 1),
        dtype=np.int64,
    )
    product = 1
    path_k_i_counter = 0
    for path_chain in path_chains_at_k:
        for last_hop in factor_last_hop:
            res = np.append(path_chain, np.asarray(last_hop, dtype=np.int64))
            rest_str = "".join([str(item) for item in res])
            extended_paths_k_plus_1[path_k_i_counter, :] = res
            product = product * path2prime_k_plus_1_hop[rest_str]
            path_k_i_counter += 1

    return product, extended_paths_k_plus_1


@nb.jit(nopython=True)
def inner_log_direct(
    value_k_hop,
    value_1_hop,
    float2path_k_hop,
    float2paths_nb_1hop,
    path2prime_k_plus_1_hop,
):

    path_chains_at_k = float2path_k_hop[value_k_hop]
    factor_last_hop = [prime for prime in float2paths_nb_1hop[value_1_hop].flatten()]
    extended_paths_k_plus_1 = np.zeros(
        (len(path_chains_at_k) * len(factor_last_hop), len(path_chains_at_k[0]) + 1),
        dtype=np.int64,
    )
    product = 0
    path_k_i_counter = 0
    for path_chain in path_chains_at_k:
        for last_hop in factor_last_hop:
            res = np.append(path_chain, np.asarray(last_hop, dtype=np.int64))
            rest_str = "".join([str(item) for item in res])
            extended_paths_k_plus_1[path_k_i_counter, :] = res
            product += np.log(path2prime_k_plus_1_hop[rest_str])
            path_k_i_counter += 1
    # if (extended_paths_k_plus_1.sum(axis=1) == 0).sum() > 0:
    #     print("WTF?")
    return product, extended_paths_k_plus_1


@nb.jit(nopython=True)
def sparse_dot2(
    matrix_left_data,
    matrix_left_indices,
    matrix_left_indptr,
    left_n_rows,
    matrix_right_data,
    matrix_right_indices,
    matrix_right_indptr,
    right_n_cols,
    number2paths_nb,
    number2paths_nb_1hop,
    path2prime_k_plus_1_hop_nb,
    semiring,
):
    """Sparse matrix multiplication matrix_left x matrix_right

    Both matrices must be in the CSR sparse format.
    Data, indices and indptr are the standard CSR representation where the column indices for row i
    are stored in indices[indptr[i]:indptr[i+1]] and their corresponding values are stored in
    data[indptr[i]:indptr[i+1]].

    Args:
        matrix_left_data (numpy.array): Non-zero value of the sparse matrix.
        matrix_left_indices (numpy.array): Column positions of non-zero values.
        matrix_left_indptr (numpy.array): Array with the count of non-zero values per row.
        matrix_right_data (numpy.array): Non-zero value of the sparse matrix.
        matrix_right_indices (numpy.array): Column positions of non-zero values.
        matrix_right_indptr (numpy.array): Array with the count of non-zero values per row.

    Returns:
        numpy.array: 2D array with the result of the matrix multiplication.
    """

    rows, cols, values = [], [], []
    k_plus_1_value_to_map = Dict.empty(
        key_type=types.float64,
        value_type=types.int64[:, :],
    )
    value = 0
    for row_left in range(left_n_rows):
        for left_i in range(
            matrix_left_indptr[row_left], matrix_left_indptr[row_left + 1]
        ):
            col_left = matrix_left_indices[left_i]
            value_left = matrix_left_data[left_i]
            for right_i in range(
                matrix_right_indptr[col_left], matrix_right_indptr[col_left + 1]
            ):
                col_right = matrix_right_indices[right_i]
                value_right = matrix_right_data[right_i]
                if "lossless" in semiring:
                    if "log" in semiring:
                        value, extended_paths_k_plus_1 = inner_log_direct(
                            value_left,
                            value_right,
                            number2paths_nb,
                            number2paths_nb_1hop,
                            path2prime_k_plus_1_hop_nb,
                        )
                    else:
                        # value = 0
                        value, extended_paths_k_plus_1 = inner_direct(
                            value_left,
                            value_right,
                            number2paths_nb,
                            number2paths_nb_1hop,
                            path2prime_k_plus_1_hop_nb,
                        )
                # else:
                #     value = value_left * value_right
                #     extended_paths_k_plus_1 = [value_left, value_right]
                k_plus_1_value_to_map[value] = extended_paths_k_plus_1

                # print(
                #     f"P^2[{row_left}, {col_right}] = {value}, generated from P^1[{row_left}, {col_left}] = {value_left},  P^1[_, {col_right}] = {value_left} Sum = {value_left+value_right}"
                # )
                # print(k_plus_1_value_to_map[value])

                rows.append(row_left)
                cols.append(col_right)
                values.append(value)
    return rows, cols, values, k_plus_1_value_to_map


def spmm(
    matrix_left: csr_array,
    matrix_right: csr_array,
    number2paths_nb,
    number2paths_nb_1hop,
    path2prime_k_plus_1_hop_nb,
    semiring,
):

    if semiring not in ["lossless", "plus_times", "lossless_log_plus"]:
        raise NotImplementedError(f"{semiring} not implemented...")

    rows, cols, values, k_plus_1_value_to_map = sparse_dot2(
        matrix_left.data,
        matrix_left.indices,
        matrix_left.indptr,
        matrix_left.shape[0],
        matrix_right.data,
        matrix_right.indices,
        matrix_right.indptr,
        matrix_right.shape[1],
        number2paths_nb,
        number2paths_nb_1hop,
        path2prime_k_plus_1_hop_nb,
        semiring=semiring,
    )

    dictionary_of_next_hop = {}
    k_plus_1_value_to_map_aggregated = {}
    for index, (row, col) in enumerate(zip(rows, cols)):
        value = values[index]
        # if "log" in semiring:
        #    value = np.log(value)
        if (row, col) not in dictionary_of_next_hop:
            dictionary_of_next_hop[(row, col)] = value
            k_plus_1_value_to_map_aggregated[value] = np.asarray(
                k_plus_1_value_to_map[value], np.int64
            )
        else:
            cur_value = dictionary_of_next_hop[(row, col)]
            if semiring == "plus_times" or semiring == "lossless_log_plus":
                dictionary_of_next_hop[(row, col)] = cur_value + value
            else:
                dictionary_of_next_hop[(row, col)] = cur_value * value
            k_plus_1_value_to_map_aggregated[
                dictionary_of_next_hop[(row, col)]
            ] = np.vstack(
                (
                    k_plus_1_value_to_map_aggregated[cur_value],
                    np.asarray(k_plus_1_value_to_map[value], np.int64),
                )
            )
    indices = list(zip(*list(dictionary_of_next_hop.keys())))
    if "log" in semiring:
        type_ = np.float64
    else:
        type_ = np.int64

    return (
        csr_array(
            (list(dictionary_of_next_hop.values()), (indices[0], indices[1])),
            shape=(matrix_left.shape[0], matrix_right.shape[1]),
            dtype=type_,
        ),
        k_plus_1_value_to_map_aggregated,
    )


def create_lossless_khops(
    df_train_orig: pd.DataFrame, max_hop: int, print_: bool = False
):
    (A_sparse, _, node2id, rel2id, broke_cause_of_sparsity,) = create_pam_matrices(
        df_train_orig,
        max_order=2,
        method="plus_times",
        use_log=False,  ############################## HARDCODED FALSE  #########
        spacing_strategy="step_1",
    )

    if print_:
        print(f"Creating Mappings..")
    mappings = create_lattice_mappings(rel2id, max_hop, False)

    number2paths = {}
    for value, paths in mappings[1]["prime2path"].items():
        number2paths[value] = paths

    number2paths_nb = Dict.empty(
        key_type=types.float64,
        value_type=types.int64[:, :],
    )
    number2paths_nb_1hop = Dict.empty(
        key_type=types.float64,
        value_type=types.int64[:, :],
    )
    for k, v in number2paths.items():
        number2paths_nb[k] = np.asarray(v, dtype=np.int64).reshape(-1, 1)
        number2paths_nb_1hop[k] = number2paths_nb[k]

    for value in A_sparse.data:
        if value not in number2paths_nb_1hop:
            number2paths_nb[value] = np.asarray(
                factorint(value, multiple=True), dtype=np.int64
            ).reshape(-1, 1)
            number2paths_nb_1hop[value] = number2paths_nb[value]

    power_A = [A_sparse]

    type_ = types.float64
    for cur_hop_index in range(max_hop - 1):
        # Prepare mappings
        if print_:
            print(f"K is {cur_hop_index + 2}")
        path2prime_k_plus_1_hop = mappings[cur_hop_index + 2]["path2prime"]

        # print(cur_hop_index + 1)
        # print(prime2path_k_hop)
        # print(path2prime_k_plus_1_hop)
        # print("\n")

        path2prime_k_plus_1_hop_nb = Dict.empty(
            key_type=types.string,
            value_type=type_,
        )

        for k, v in path2prime_k_plus_1_hop.items():
            path2prime_k_plus_1_hop_nb["".join([str(k_) for k_ in k])] = v

        updated_A, new_paths = spmm(
            power_A[-1],
            power_A[0],
            number2paths_nb,
            number2paths_nb_1hop,
            path2prime_k_plus_1_hop_nb,
            "lossless_log_plus",
        )

        mappings[cur_hop_index + 2] = {"path2prime": {}, "prime2path": {}}
        for float_, paths_arr in new_paths.items():
            number2paths_nb[float_] = np.asarray(paths_arr, dtype=np.int64)
            path_list = paths_arr.tolist()

            tupled_paths = tuple([tuple(sub) for sub in path_list])
            mappings[cur_hop_index + 2]["path2prime"][tupled_paths] = float_
            mappings[cur_hop_index + 2]["prime2path"][float_] = tupled_paths

        num_overflow = (updated_A.data < 0).sum()
        num_inf = np.isinf(updated_A.data).sum()
        updated_A.data[(updated_A.data < 0)] = 0
        updated_A.data[np.isinf(updated_A.data)] = 0
        updated_A.eliminate_zeros()
        if num_overflow > 0 or num_inf > 0:
            # pass
            # raise ArithmeticError(f"Created {num_overflow} overflows and {num_inf} infinities during calculations")

            print(
                f"Hop {cur_hop_index + 2}: Created {num_overflow} overflows and {num_inf} infinities during calculations"
            )
        from utils import get_sparsity

        sparsity = get_sparsity(updated_A)
        if print_:
            print(f"Sparsity {cur_hop_index + 2}-hop: {sparsity:.2f} %")

        power_A.append(updated_A)

    return A_sparse, power_A, node2id, rel2id, broke_cause_of_sparsity, mappings


################################################## HARD-CODED FASLE USE_LOG ########################################3

if __name__ == "__main__":

    import time

    from data_loading import load_data

    project_name = "test"
    add_inverse_edges = "NO"
    path = "../data/dummy_data"
    max_hop = 3
    print_ = True

    df_train_orig, df_train, df_eval, df_test, already_seen_triples = load_data(
        path,
        project_name=project_name,
        add_inverse_edges=add_inverse_edges,
        sep=",",
        skiprows=0,
    )

    total_s = time.time()
    (
        A_sparse,
        power_A,
        node2id,
        rel2id,
        broke_cause_of_sparsity,
        mappings,
    ) = create_lossless_khops(df_train_orig, max_hop, print_)

    (
        pam_1hop_lossless,
        power_A,
        node2id,
        rel2id,
        broke_cause_of_sparsity,
        mappings,
    ) = create_lossless_khops(df_train_orig, max_hop=max_hop)

    time_taken = time.time() - total_s
    print(f"Overall {time_taken:.2f} seconds ({time_taken/60:.2f} mins)")



import math
import string

import numpy as np
import pandas as pd
import tqdm
from scipy.sparse import csr_array

from utils import get_primefactors_multiplicity


def noisy_or_aggregator(confidences: list):
    """Noisy or aggregator according to AnyBURL
    for aggregating multiple rule confidences in one score.
    This is a strategy to merge the confidence of multiple
    rules that were triggered for the same prediction into
    one score.

    Args:
        confidences (list): List of scores.

    Returns:
        float: The aggregated score.
    """
    return 1 - np.prod([1 - confindence for confindence in confidences])


def generate_rule_string(
    chain_of_relations: list[str], head_rel: str
) -> tuple[str, str, str]:
    """Helper function to create the rule

    Args:
        chain_of_relations (list[str]): The chain of relations followed.
        head_rel (str): The relation implied in the head.

    Returns:
        tuple[str, str, str]: The string of the rule, the first letter (A) and the last
        letter (B). The last two values are there to help change the final result.
    """

    possible_letters = string.ascii_uppercase
    first_letter = possible_letters[0]
    rule_str = ""
    for i_rel, relation in enumerate(chain_of_relations):
        next_letter = possible_letters[possible_letters.index(first_letter) + 1]
        rule_str += f"{relation}( {first_letter} , {next_letter} ) ^ "
        first_letter = next_letter

    rule_str = rule_str[:-2]
    head = f"-> {head_rel}( {possible_letters[0]} , {next_letter} )"

    rule_str += head
    return rule_str, possible_letters[0], next_letter


def map_dataframe(
    df: pd.DataFrame,
    node2id: dict[str, int],
    rel2id: dict[str, int],
    dropna: bool = True,
) -> pd.DataFrame:
    """Helper function that maps node and relation names to integers.

    Args:
        df (pd.DataFrame): The original dataframe
        node2id (dict[str, int]): The dictionary mapping nodes to their ids e.g. "Ent1"->1.
        rel2id (dict[str, int]): The dictionary mapping relations to their ids e.g. "rel1"->3.
        dropna (bool): If True, drops the unmapped rows.

    Returns:
        pd.DataFrame: The mapped dataframe. If any could not be mapped, are simply dropped.
    """
    df_mapped = df.copy()
    df_mapped["rel"] = df["rel"].map(rel2id)
    df_mapped["head"] = df["head"].map(node2id)
    df_mapped["tail"] = df["tail"].map(node2id)

    if dropna:
        # print(f"Length before dropping nan", len(df_mapped))
        df_mapped.dropna(inplace=True)
        # print(f"Length after dropping nan", len(df_mapped))
    df_mapped["head"] = df_mapped["head"].astype(int)
    df_mapped["tail"] = df_mapped["tail"].astype(int)
    return df_mapped


def predict_tail_with_explanations(
    df_test: pd.DataFrame,
    all_rules_df: pd.DataFrame,
    pam_powers: list[csr_array],
    node2id: dict[str, int],
    rel2id: dict[str, int],
    rank_rules_by_: str = "score",
) -> None:
    """_summary_

    Args:
        df_test (pd.DataFrame): The dataframe with columns ['head', 'rel', 'tail'] containing the test triples.
        all_rules_df (pd.DataFrame): The dataframe with the rules as generated by rule_generation.py.
        pam_powers (list[csr_array]): The PAM powers to check for triggered rules.
        node2id (dict[str, int]): The dictionary mapping nodes to their ids e.g. "Ent1"->1.
        rel2id (dict[str, int]): The dictionary mapping relations to their ids e.g. "rel1"->3
        rank_rules_by_ (str): The attribute over which we decide the importance of the rule
    Returns:
        Nothing. Simply prints each prediction and exits
    """
    # Map each node to the corresponding PAM index ID
    # And each relationship to its prime number
    df_test_mapped = map_dataframe(df_test, node2id, rel2id)

    # Create the inverse dicts as well for printing purposes
    id2rel = dict(zip(rel2id.values(), rel2id.keys()))
    id2node = dict(zip(node2id.values(), node2id.keys()))

    # Iterate over each query
    for test_index, test_row in df_test_mapped.iterrows():

        probable_predictions = 0
        query_head = test_row["head"]
        query_rel = test_row["rel"]
        query_tail = test_row["tail"]
        print(
            f"Query ({test_index}): {id2node[query_head]} - {id2rel[query_rel]} - {id2node[query_tail]}"
        )

        # Keep only the rules that imply this relationship
        query_rules_all_hops = all_rules_df[all_rules_df["head_rel"] == query_rel]
        # Group them by hop-level (i.e. all rules of length 2, all rules of length 3 etc.)
        hop_level_to_rule_body_values = (
            query_rules_all_hops.groupby("hop")["body"].aggregate(list).to_dict()
        )
        # Iterate at each hop-level to find possible matches
        for k_hop_level, body_values in hop_level_to_rule_body_values.items():

            # Keep track of the confidence that each rule has
            k_hop_rules_conf = query_rules_all_hops[
                query_rules_all_hops["hop"] == k_hop_level
            ][["body", rank_rules_by_]]
            k_hop_rules_conf_dict = dict(
                zip(k_hop_rules_conf["body"], k_hop_rules_conf[rank_rules_by_])
            )
            # Find the corresponding row for this query head from the appropriate
            # k-hop matrix
            k_hop_pam_head_chains = pam_powers[k_hop_level - 1][[query_head]]
            # Find where the nnz entries of this row, match any of the rules that
            # we have recongized
            probable_tail_entities_indices = np.where(
                np.in1d(k_hop_pam_head_chains.data, body_values)
            )[0]
            # If there exist such cases, then some rules are triggered and these
            # lead us to predictions
            if probable_tail_entities_indices.shape[0] > 0:
                # Keep note of the possible tails
                probable_tail_entities_nodes = k_hop_pam_head_chains.indices[
                    probable_tail_entities_indices
                ]
                # Keep note of the rules that generated each tail
                probable_tail_triggered_rules = k_hop_pam_head_chains.data[
                    probable_tail_entities_indices
                ]
                # Keep note of the confidence of each rule
                probable_tail_triggered_rules_confidences = np.array(
                    [
                        k_hop_rules_conf_dict[rule_value]
                        for rule_value in probable_tail_triggered_rules
                    ]
                )

                # Get ready for printing
                # Re-create the original relation that needs to be predicted
                head_rel = id2rel[query_rel]
                # Iterate over each possible prediction and print it
                for pred_index, predicted_node in enumerate(
                    probable_tail_entities_nodes
                ):
                    # Check whether the prediction was correct
                    result = (
                        "Correct Match"
                        if predicted_node == query_tail
                        else "Wrong Match"
                    )
                    # Keep track of the body of the rule
                    rule_body = probable_tail_triggered_rules[pred_index]
                    # Keep track of the confidence of the rule
                    rule_score = probable_tail_triggered_rules_confidences[pred_index]
                    # Check whether we can deconstruct the rule to the constituent steps,
                    # If yes do so
                    # Else leave the whole rule body as a number as is
                    if math.ceil(rule_body) == rule_body:
                        # If we can factorize the rules, recreate the semantic chain
                        try:
                            chain = [
                                id2rel[prime_rel]
                                for prime_rel in get_primefactors_multiplicity(
                                    int(rule_body)
                                )
                            ]
                            # Generate the rule string
                            rule_str, first_letter, last_letter = generate_rule_string(
                                chain, head_rel
                            )
                            # Relplace the starting and ending letters with the appropriate
                            # Query values
                            rule_str = rule_str.replace(
                                f" {first_letter} ", id2node[query_head]
                            ).replace(f" {last_letter} ", id2node[query_tail])
                        except KeyError:
                            print(f"Could not decompose chain {rule_body}")
                            rule_str = f"{rule_body}({id2node[query_head]}, {id2node[query_tail]})->{id2rel[query_rel]}({id2node[query_head]}, {id2node[query_tail]}))"
                    else:
                        print(f"Could not decompose chain {rule_body}")
                        rule_str = f"{rule_body}({id2node[query_head]}, {id2node[query_tail]})->{id2rel[query_rel]}({id2node[query_head]}, {id2node[query_tail]}))"
                    # Print each prediction
                    print(f"({result}) {rule_score:.4f} : {rule_str}\t")
                    print("\n\n")
                    probable_predictions += 1
        if probable_predictions == 0:
            print(f"No rules triggered for this query...\n\n")
    return None


def predict_tail_with_ranks(
    df_test: pd.DataFrame,
    all_rules_df: pd.DataFrame,
    pam_powers: list[csr_array],
    node2id: dict[str, int],
    rel2id: dict[str, int],
    rank_rules_by_: str = "score",
    aggregate_rules_by_str: str = "max",
) -> pd.DataFrame:
    """_summary_

    Args:
        df_test (pd.DataFrame): The dataframe with columns ['head', 'rel', 'tail'] containing the test triples.
        all_rules_df (pd.DataFrame): The dataframe with the rules as generated by rule_generation.py.
        pam_powers (list[csr_array]): The PAM powers to check for triggered rules.
        node2id (dict[str, int]): The dictionary mapping nodes to their ids e.g. "Ent1"->1.
        rel2id (dict[str, int]): The dictionary mapping relations to their ids e.g. "rel1"->3
        rank_rules_by_ (str): The attribute over which we decide the importance of the rule
        aggregate_rules_by_str (str): The aggregation function. Currently either 'max' or 'noisy_or'
    Returns:
        df_results (pd.DataFrame): The predictions made. One row per test query.
    """

    if aggregate_rules_by_str == "max":
        aggregate_rules_by_func = max
    elif aggregate_rules_by_str == "noisy_or":
        aggregate_rules_by_func = noisy_or_aggregator
    else:
        raise AttributeError(
            f"Aggregate function string {aggregate_rules_by_str} not recognized.."
        )
    # Map each node to the corresponding PAM index ID
    # And each relationship to its prime number
    df_test_mapped = map_dataframe(df_test, node2id, rel2id)

    # Iterate over each query
    results = []
    for _, test_row in tqdm.tqdm(
        df_test_mapped.iterrows(), total=df_test_mapped.shape[0]
    ):

        probable_tails = []
        query_head = test_row["head"]
        query_rel = test_row["rel"]
        query_tail = test_row["tail"]

        # Keep only the rules that imply this relationship
        query_rules_all_hops = all_rules_df[all_rules_df["head_rel"] == query_rel]
        # Group them by hop-level (i.e. all rules of length 2, all rules of length 3 etc.)
        hop_level_to_rule_body_values = (
            query_rules_all_hops.groupby("hop")["body"].aggregate(list).to_dict()
        )
        # Iterate at each hop-level to find possible matches
        for k_hop_level, body_values in hop_level_to_rule_body_values.items():

            # Keep track of the confidence that each rule has
            k_hop_rules_conf = query_rules_all_hops[
                query_rules_all_hops["hop"] == k_hop_level
            ][["body", rank_rules_by_]]
            k_hop_rules_conf_dict = dict(
                zip(k_hop_rules_conf["body"], k_hop_rules_conf[rank_rules_by_])
            )
            # Find the corresponding row for this query head from the appropriate
            # k-hop matrix
            k_hop_pam_head_chains = pam_powers[k_hop_level - 1][[query_head]]
            # Find where the nnz entries of this row, match any of the rules that
            # we have recongized
            probable_tail_entities_indices = np.where(
                np.in1d(k_hop_pam_head_chains.data, body_values)
            )[0]
            # If there exist such cases, then some rules are triggered and these
            # lead us to predictions
            if probable_tail_entities_indices.shape[0] > 0:
                # Keep note of the possible tails
                probable_tail_entities_nodes = k_hop_pam_head_chains.indices[
                    probable_tail_entities_indices
                ]
                # Keep note of the rules that generated each tail
                probable_tail_triggered_rules = k_hop_pam_head_chains.data[
                    probable_tail_entities_indices
                ]
                # Keep note of the confidence of each rule
                probable_tail_triggered_rules_confidences = np.array(
                    [
                        k_hop_rules_conf_dict[rule_value]
                        for rule_value in probable_tail_triggered_rules
                    ]
                )
                probable_tail_hop_levels = np.array(
                    [k_hop_level] * len(probable_tail_entities_nodes)
                ).reshape(-1, 1)
                probable_tails.append(
                    np.hstack(
                        (
                            probable_tail_entities_nodes.reshape(-1, 1),
                            probable_tail_triggered_rules.reshape(-1, 1),
                            probable_tail_triggered_rules_confidences.reshape(-1, 1),
                            probable_tail_hop_levels,
                        )
                    )
                )
        if probable_tails:
            predictions = pd.DataFrame(
                np.vstack(probable_tails),
                columns=["Tail", "Rule_Triggered", "Rule_Confidence", "Hop_Level"],
            )
            predictions["Tail"] = predictions["Tail"].astype(int)
            sorted_predictions = (
                predictions.groupby(["Tail"])["Rule_Confidence"]
                .apply(aggregate_rules_by_func)
                .sort_values(ascending=False)
            )
            best_candidates = sorted_predictions.index.values.astype(int).tolist()
            best_candidates_scores = sorted_predictions.values.tolist()
        else:
            best_candidates = []
            best_candidates_scores = []
        test_row["predictions"] = best_candidates
        test_row["predictions_scores"] = best_candidates_scores
        try:
            test_row["rank_correct"] = best_candidates.index(query_tail) + 1
        except ValueError:
            test_row["rank_correct"] = 0
        results.append(test_row)
    df_results = pd.DataFrame(results)
    return df_results


if __name__ == "__main__":

    import time

    from data_loading import load_data
    from pam_creation import create_pam_matrices
    from rule_generation import create_ruleset
    from utils import calculate_hits_at_k

    # Data Loader details
    # Path to data folder containing train.txt and test.txt
    path = "../data/dummy_data"
    # Check load_data and change it according to the particularities of your dataset
    # e.g. the separator is '\t' instead of ','
    project_name = "test"
    # Whether to augment the graph with inverse edges
    add_inverse_edges = "NO"
    # Whether to map primes to logs. This is used for numerical overflows in
    # big data. It messes with explainability.
    use_log = False
    # What is the max k-hop PAM to create
    max_num_hops = 5
    # How far apart will the primes be
    spacing_strategy = "step_10"
    # What is the multiplication strategy
    method = "plus_times"
    # During inference, rank rules according to what attribute
    rank_rules_by_ = "score"
    # During inference, aggregate multiple rules into one by keeping their best result
    aggregate_rules_by_str = "max"

    time_s = time.time()
    df_train_orig, df_train, df_eval, df_test, already_seen_triples = load_data(
        path, project_name=project_name, add_inverse_edges=add_inverse_edges, sep=","
    )
    print(f"\nLoaded Data, will create PAMs... \n")

    (
        pam_1hop_lossless,
        pam_powers,
        node2id,
        rel2id,
        broke_cause_of_sparsity,
    ) = create_pam_matrices(
        df_train,
        max_order=max_num_hops,
        method=method,
        use_log=use_log,
        spacing_strategy=spacing_strategy,
        break_with_sparsity_threshold=-1,
    )
    print(f"\nCreated PAMs, will generate rules... \n")

    all_rules_df = create_ruleset(
        pam_1hop_lossless, pam_powers, use_log=use_log, max_num_hops=-1
    )
    print(f"\nCreated {all_rules_df.shape[0]} rules, will generate predictions...  \n")

    k_hop_pams = [pam_1hop_lossless] + pam_powers[1:]

    predict_tail_with_explanations(
        df_test, all_rules_df, k_hop_pams, node2id, rel2id, rank_rules_by_="score"
    )

    df_results = predict_tail_with_ranks(
        df_test,
        all_rules_df,
        k_hop_pams,
        node2id,
        rel2id,
        rank_rules_by_="score",
        aggregate_rules_by_str="max",
    )
    _ = calculate_hits_at_k(df_results, print_=True)

    time_total = time.time() - time_s
    print(
        f"\nThe whole process took: {time_total:.0f} sec. ({time_total/60:.2f} mins).."
    )



import functools
import random

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from sympy import factorint, primefactors


def set_all_seeds(seed: int = 0):
    """Fix random seeds

    Args:
        seed (int): Random seed
    """

    random.seed(seed)
    np.random.seed(seed)
    return 1


def get_sparsity(A: csr_matrix) -> float:
    """Calculate sparsity % of scipy sparse matrix.

    Args:
        A (scipy.sparse): Scipy sparse matrix

    Returns:
        (float)): Sparsity as a float
    """

    return 100 * (1 - A.nnz / (A.shape[0] ** 2))


@functools.lru_cache(maxsize=None)
def get_primefactors(value: float) -> list[int]:
    """Wrapper functiom that gets a value and returns the list
       of prime factors of the value. It is used as a wrapper around
       primefactors in ordet ot use memoization with cache for speed.

    Args:
        value (float): The float value to decompose

    Returns:
        Tuple[int]: A list of the unique prime factors
    """
    return primefactors(value)


@functools.lru_cache(maxsize=None)
def get_primefactors_multiplicity(value: float) -> list[int]:
    """Wrapper function that gets a value and returns the list
       of prime factors of the value with multiplicity. It is used as a wrapper around
       factorint in ordet ot use memoization with cache for speed.

    Args:
        value (float): The float value to decompose

    Returns:
        list[int]: A list of the (multiple maybe) prime factors
    """

    total = []
    for factor, multiplicity in factorint(value).items():
        total.extend([factor] * multiplicity)
    return total


def calculate_hits_at_k(
    results: pd.DataFrame, wanted_ranks: list[int] = [1, 3, 5, 10], print_: bool = False
) -> dict[str, float]:
    """Helper function to print results.
    The given dataframe has one column named 'rank_correct' ranging from 0 to inf
    denoting the rank of the correct target entity in the predictions for each query (row).
    If the correct target was not found it is 0.
    Will also print the upper bound, meaning if we had an ORACLE selector to provide us with the
    correct target each time irrelevant of the actual rank of the target, what would our score be.
    This essentially is the percentage of test samples where the correct target was found somewhere in
    the predictions made.

    Args:
        results (pd.DataFrame): Dataframe with the predictions per test query.
        wanted_ranks (list[int], optional): Ranks at which we calculate the values. Defaults to [1, 3, 5, 10].
        print_ (bool, optional): Whether to print them as well. Defaults to False.

    Returns:
        dict[str,float]: The ranks in dict
    """
    hits_at_k = {}
    num_test_samples = results.shape[0]
    for k in wanted_ranks:
        hits_at_k[k] = (
            results[
                (results["rank_correct"] > 0) & (results["rank_correct"] <= k)
            ].shape[0]
            / num_test_samples
        )
    hits_at_k["UpperBound"] = (
        num_test_samples - results[(results["rank_correct"] == 0)].shape[0]
    ) / num_test_samples
    if print_:
        for hit_k, perc_hit in hits_at_k.items():
            print(f"Tail Hits@{hit_k}: {100*perc_hit:.2f} %")
    return hits_at_k



import os

import pandas as pd
from sympy import nextprime, primefactors


def load_csv(
    path_to_file: str,
    add_inverse_edges: str = "NO",
    pandas_sep=",",
    pandas_first_row_header=False,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Helper function that loads the collection of triples in pd.DataFrames and returns them.
    It expects a collection of triples as .csv where each triple is in one line. E.g.
    entity1, relation1, entity2,
    entity1, relation2, entity3,
    ...
    If there is a header first row, please pass pandas_first_row_header=True
    Args:
        path_to_file (str): path to .csv co
        add_inverse_edges (str, optional):  Whether to add the inverse edges.
        Possible values "YES", "YES__INV", "NO". Defaults to "NO".
        If "YES" will simply add the inverse edges. Eg.
        If "entity1, relation1, entity2" exists it will add "entity2, relation1, entity1"
        If "YES__INV" it will add the inverse edges with a suffix. Eg.
        If "entity1, relation1, entity2" exists it will add "entity2, relation1__INV, entity1"
        If "NO" it will add no new triples.
        pandas_sep (str, optional): the separator to use in pd.read_csv,
        when loading the data. Defaults to ",".
        pandas_first_row_header (bool, optional): whether the first row of the file
        contains a header for the data. Defaults to False.
    Returns:
        Tuple[pd.DataFrame, pd.DataFrame]: the triples loaded in a pandas. We return
        both the original triple-set, as well as the (possibly) augmented one
        (i.e. with inverse edges)
    """

    if pandas_first_row_header:
        header_row = 0
    else:
        header_row = None

    df_train = pd.read_csv(
        path_to_file,
        sep=pandas_sep,
        header=header_row,
        dtype="str",
    )
    df_train.columns = ["head", "rel", "tail"]  # type: ignore
    df_train_orig = df_train.copy()
    if "YES" in add_inverse_edges:
        print(f"Will add the inverse train edges as well..")
        df_train["rel"] = df_train["rel"].astype(str)
        df_train_inv = df_train.copy()
        df_train_inv["head"] = df_train["tail"]
        df_train_inv["tail"] = df_train["head"]
        if add_inverse_edges == "YES__INV":
            df_train_inv["rel"] = df_train["rel"] + "__INV"
        df_train = pd.concat((df_train, df_train_inv))
    return df_train_orig, df_train


def load_data(
    path_to_folder: str,
    project_name: str,
    add_inverse_edges: str = "NO",
    skiprows: int = 0,
    sep: str = "\t",
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, set]:
    """
    Helper function that loads the data in pd.DataFrames and returns them.
    No header row is expected here. It follows the same convention as load_csv dataset.
    Args:
        path_to_folder (str): path to folder with train.txt, valid.txt, test.txt
        project_name (str): name of the project
        add_inverse_edges (str, optional):  Whether to add the inverse edges.
        Possible values "YES", "YES__INV", "NO". Defaults to "NO".
        skiprows (int, optional): Number of rows to skip. Pleas provide if new dataset.
        Default is 0.
        sep (str, optional): Separator for read_csv. Please provide if new dataset.
        Default is '\t'.

    Returns:
        Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, set]: [description]
    """
    PROJECT_DETAILS = {
        "lc-neo4j": {"skiprows": 1, "sep": "\t"},
        "codex-s": {"skiprows": 0, "sep": "\t"},
        "codex-l": {"skiprows": 0, "sep": "\t"},
        "WN18RR": {"skiprows": 0, "sep": "\t"},
        "YAGO3-10-DR": {"skiprows": 0, "sep": "\t"},
        "YAGO3-10": {"skiprows": 0, "sep": "\t"},
        "FB15k-237": {"skiprows": 0, "sep": "\t"},
        "NELL995": {"skiprows": 0, "sep": "\t"},
        "DDB14": {"skiprows": 0, "sep": "\t"},
    }

    # Try out with generic skiprows=0 and sep='\t'
    if not project_name in PROJECT_DETAILS:
        PROJECT_DETAILS[project_name] = {"skiprows": skiprows, "sep": sep}

    df_train = pd.read_csv(
        os.path.join(path_to_folder, "train.txt"),
        sep=PROJECT_DETAILS[project_name]["sep"],
        header=None,
        dtype="str",
        skiprows=PROJECT_DETAILS[project_name]["skiprows"],
    )
    df_eval = []
    df_test = []
    df_train.columns = ["head", "rel", "tail"]  # type: ignore
    df_train_orig = df_train.copy()
    if "YES" in add_inverse_edges:
        print(f"Will add the inverse train edges as well..")
        df_train["rel"] = df_train["rel"].astype(str)
        df_train_inv = df_train.copy()
        df_train_inv["head"] = df_train["tail"]
        df_train_inv["tail"] = df_train["head"]
        if add_inverse_edges == "YES__INV":
            df_train_inv["rel"] = df_train["rel"] + "__INV"
        df_train = df_train.append(df_train_inv)
    if project_name in ["lc-neo4j"]:
        df_eval = pd.DataFrame()
        df_test = pd.DataFrame()
        already_seen_triples = set(df_train.to_records(index=False).tolist())
    else:
        try:
            df_eval = pd.read_csv(
                os.path.join(path_to_folder, "valid.txt"),
                sep=PROJECT_DETAILS[project_name]["sep"],
                header=None,
                dtype="str",
                skiprows=PROJECT_DETAILS[project_name]["skiprows"],
            )
            df_eval.columns = ["head", "rel", "tail"]  # type: ignore
        except FileNotFoundError:
            print(
                f"No valid.txt found in {path_to_folder}... df_eval will contain the train data.."
            )
            df_eval = df_train.copy()
        df_test = pd.read_csv(
            os.path.join(path_to_folder, "test.txt"),
            sep=PROJECT_DETAILS[project_name]["sep"],
            header=None,
            dtype="str",
            skiprows=PROJECT_DETAILS[project_name]["skiprows"],
        )
        df_test.columns = ["head", "rel", "tail"]  # type: ignore
        if "YAGO" in project_name:
            for cur_df in [df_train, df_eval, df_test]:
                for col in cur_df.columns:
                    cur_df[col] = cur_df[col]  # + "_YAGO"

        already_seen_triples = set(
            df_train.to_records(index=False).tolist()
            + df_eval.to_records(index=False).tolist()
        )
    print(f"Total: {len(already_seen_triples)} triples in train + eval!)")
    print(f"In train: {len(df_train)}")
    print(f"In valid: {len(df_eval)}")
    print(f"In test: {len(df_test)}")
    return df_train_orig, df_train, df_eval, df_test, already_seen_triples


if __name__ == "__main__":
    path = "../data/dummy_data/train.txt"
    df_train_orig, df_train = load_csv(path, add_inverse_edges="YES")
    print(df_train_orig)
    assert df_train_orig.shape[0] * 2 == df_train.shape[0]



from itertools import product

import numpy as np
import pandas as pd
import tqdm
from scipy.sparse import csr_array


def generate_path_features(
    pam_powers: list[csr_array],
    pairs: list[tuple[int, int]] = [],
) -> pd.DataFrame:
    """ "
    Function to generate path features for the pairs given.
    (Or all pairs if no list of pairs is given).

    Args:
        pam_powers (list[csr_array]): List of sparse arrays.
        pairs (list[tuple[int, int]], optional): Pair-ids for which to generate the features. If none given
        all are generated. Defaults to [].

    Returns:
        pd.DataFrame:
    """

    num_nodes = pam_powers[0].shape[0]
    if len(pairs) == 0:
        pairs = [
            item for item in product(list(range(num_nodes)), list(range(num_nodes)))
        ]
    features = []
    for pair in pairs:
        cur_features = []
        for k_hop_pam in pam_powers:
            cur_features.append(k_hop_pam[pair[0], pair[1]])
        features.append(np.array(cur_features))
    feats = np.array(features)
    feats = pd.DataFrame(feats, columns=[f"val@{k+1}" for k in range(len(pam_powers))])
    feats["pairs"] = pairs
    return feats


if __name__ == "__main__":
    from data_loading import load_data
    from pam_creation import create_pam_matrices

    path = "../data/dummy_data"

    df_train_orig, df_train, df_eval, df_test, already_seen_triples = load_data(
        path, project_name="test", add_inverse_edges="NO", sep=","
    )
    (
        pam_1hop_lossless,
        pam_powers,
        node2id,
        rel2id,
        broke_cause_of_sparsity,
    ) = create_pam_matrices(
        df_train,
        max_order=5,
        method="plus_times",
        use_log=False,
        eliminate_diagonal=False,
        spacing_strategy="step_1",
    )
    path_feats = generate_path_features(pam_powers, pairs=[])
    print(path_feats)
